Documentation.addTranslations({"locale": "pl", "messages": {"%(filename)s &#8212; %(docstitle)s": "%(filename)s &#8212; %(docstitle)s", "&#169; <a href=\"%(path)s\">Copyright</a> %(copyright)s.": "&#169; <a href=\"%(path)s\">Copyright</a> %(copyright)s.", "&#169; Copyright %(copyright)s.": "&#169; Copyright %(copyright)s.", ", in ": ", w ", "About these documents": "O tych dokumentach", "Automatically generated list of changes in version %(version)s": "Automatycznie wygenerowana lista zmian w wersji %(version)s", "C API changes": "Zmiany w C API", "Changes in Version %(version)s &#8212; %(docstitle)s": "Zmiany w wersji %(version)s &#8212; %(docstitle)s", "Collapse sidebar": "Zwi\u0144 pasek boczny", "Complete Table of Contents": "Kompletny spis tre\u015bci", "Contents": "Tre\u015b\u0107", "Copyright": "Copyright", "Created using <a href=\"http://sphinx-doc.org/\">Sphinx</a> %(sphinx_version)s.": "Utworzone przy pomocy <a href=\"http://sphinx-doc.org/\">Sphinx</a>'a %(sphinx_version)s.", "Expand sidebar": "Rozwi\u0144 pasek boczny", "From here you can search these documents. Enter your search\n    words into the box below and click \"search\". Note that the search\n    function will automatically search for all of the words. Pages\n    containing fewer words won't appear in the result list.": "St\u0105d mo\u017cesz przeszuka\u0107 dokumentacj\u0119. Wprowad\u017a szukane\n    s\u0142owa w poni\u017cszym okienku i kliknij \"Szukaj\". Zwr\u00f3\u0107 uwag\u0119, \u017ce\n    funkcja szukaj\u0105ca b\u0119dzie automatycznie szuka\u0142a wszystkich s\u0142\u00f3w. Strony\n    nie zawieraj\u0105ce wszystkich wpisanych s\u0142\u00f3w nie znajd\u0105 si\u0119 na wynikowej li\u015bcie.", "Full index on one page": "Ca\u0142y indeks na jednej stronie", "General Index": "Indeks og\u00f3lny", "Global Module Index": "Globalny indeks modu\u0142\u00f3w", "Go": "Szukaj", "Hide Search Matches": "Ukryj wyniki wyszukiwania", "Index": "Indeks", "Index &ndash; %(key)s": "Indeks &ndash; %(key)s", "Index pages by letter": "Strony indeksu alfabetycznie", "Indices and tables:": "Indeksy i tablice:", "Last updated on %(last_updated)s.": "Ostatnia modyfikacja %(last_updated)s.", "Library changes": "Zmiany w bibliotekach", "Navigation": "Nawigacja", "Next topic": "Nast\u0119pny temat", "Other changes": "Inne zmiany", "Overview": "Przegl\u0105d", "Permalink to this definition": "Sta\u0142y odno\u015bnik do tej definicji", "Permalink to this headline": "Sta\u0142y odno\u015bnik do tego nag\u0142\u00f3wka", "Please activate JavaScript to enable the search\n    functionality.": "Aby umo\u017cliwi\u0107 wyszukiwanie, prosz\u0119 w\u0142\u0105czy\u0107 JavaScript.", "Preparing search...": "Inicjalizacja wyszukiwania...", "Previous topic": "Poprzedni temat", "Quick search": "Szybkie wyszukiwanie", "Search": "Szukaj", "Search Page": "Wyszukiwanie", "Search Results": "Wyniki wyszukiwania", "Search finished, found %s page(s) matching the search query.": "Wyszukiwanie zako\u0144czone. Liczba znalezionych stron pasuj\u0105cych do zapytania: %s.", "Search within %(docstitle)s": "Szukaj po\u015br\u00f3d %(docstitle)s", "Searching": "Wyszukiwanie", "Show Source": "Poka\u017c \u017ar\u00f3d\u0142o", "Table Of Contents": "Spis tre\u015bci", "This Page": "Ta strona", "Welcome! This is": "Witaj! To jest", "Your search did not match any documents. Please make sure that all words are spelled correctly and that you've selected enough categories.": "Twoje wyszukiwanie nie da\u0142o \u017cadnych wynik\u00f3w. Upewnij si\u0119, \u017ce wszystkie s\u0142owa s\u0105 wpisane prawid\u0142owo i \u017ce wybra\u0142e\u015b dostateczn\u0105 ilo\u015b\u0107 kategorii.", "all functions, classes, terms": "wszystkie funkcje, klasy, terminy", "can be huge": "mo\u017ce by\u0107 ogromny", "last updated": "ostatnio aktualizowana", "lists all sections and subsections": "wszystkie rozdzia\u0142y i podrozdzia\u0142y", "next chapter": "nast\u0119pny rozdzia\u0142", "previous chapter": "poprzedni rozdzia\u0142", "quick access to all modules": "szybki dost\u0119p do wszystkich modu\u0142\u00f3w", "search": "szukaj", "search this documentation": "przeszukaj t\u0119 dokumentacj\u0119", "the documentation for": "dokumentacja do"}, "plural_expr": "(n==1 ? 0 : (n%10>=2 && n%10<=4) && (n%100<12 || n%100>=14) ? 1 : n!=1 && (n%10>=0 && n%10<=1) || (n%10>=5 && n%10<=9) || (n%100>=12 && n%100<=14) ? 2 : 3)"});